/*
 * Source File < bgiutl.c >
 *
 * BGI Utilities Functions: Borland Graphics Interface for Xlib
 *
 * Copyleft (c) 1993 Tokyo Denki University, Taiji Yamada
 */
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "graphix.h"
#include "xbgiutl.h"
#include "bgiutl.h"

#define minmax(a,x,b)  ((((a)<=(x))&&((x)<=(b)))?(x):(((x)<(a))?(a):(b)))
#define Xx(x)          (XBGI->viewport.left+(x))
#define Xy(y)          (XBGI->viewport.top +(y))
#define xX(x)          ((x)-XBGI->viewport.left)
#define yX(y)          ((y)-XBGI->viewport.top )
#define CTRL(c)        ((c)-'a'+1)

#define HemBrim        (DecoHemBits+DecoBrimBits)
#define Hem(size)      (DecoHemBits+(size)+DecoHemBits)
#define Brim(size)     (DecoBrimBits+(size)+DecoBrimBits)
#define TextHeight(fs) (fs->max_bounds.ascent+fs->max_bounds.descent)
#define TextAscent(fs) (fs->max_bounds.ascent)

extern xbgi *XBGI;

void bgivprintf( char *title, int textjust, char *fmt, va_list argptr )
{
  Window rwin, twin, pwin;
  XEvent ev;
  int i, mwidth, twidth, theight, pwidth, pheight;
  int x, y;
  
  xbgiGetSizePrintf( XBGI->fs, &pwidth, &pheight, title );
  xbgiGetSizeVPrintf( XBGI->fs, &twidth, &theight, fmt, argptr );
  mwidth = max( twidth, pwidth );
  x = XBGI->key_press_ev.xkey.x_root;
  y = XBGI->key_press_ev.xkey.y_root;
  x = x - Hem(mwidth)/2;
  y = y - Hem(theight)/2;
  rwin = XCreateSimpleWindow( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
			     x, y, Hem(Hem(mwidth)), Hem(Hem(pheight)+Hem(theight)), 0,
			     XBGI->fg, XBGI->bg );
  XSelectInput( XBGI->dpy, rwin, ExposureMask|ButtonPressMask );
  twin = XCreateSimpleWindow( XBGI->dpy, rwin, 
			     DecoHemBits, DecoHemBits, Hem(mwidth), Hem(pheight), 0,
			     XBGI->fg, XBGI->bg );
  pwin = XCreateSimpleWindow( XBGI->dpy, rwin, 
			     DecoHemBits, DecoHemBits+Hem(pheight), 
			     Hem(mwidth), Hem(theight), 0, XBGI->fg, XBGI->bg );
  XMapWindow( XBGI->dpy, rwin );
  XMapSubwindows( XBGI->dpy, rwin );
  XSetTransientForHint( XBGI->dpy, rwin, XBGI->root );

  XFlush( XBGI->dpy );
  while (1){
    XNextEvent( XBGI->dpy, &ev );
    switch ( ev.type ){
    case Expose:
      xbgiDecorateWindow( XBGI->dpy, rwin, XBGI->gc,
			 XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
      xbgiDecorateWindow( XBGI->dpy, twin, XBGI->gc,
			 XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
      xbgiPrintf( XBGI->dpy, twin, XBGI->gc, XBGI->fs, CENTER_TEXT, "%s", title );
      xbgiDecorateWindow( XBGI->dpy, pwin, XBGI->gc,
			 XBGI->fg, XBGI->lc, XBGI->dc, DecoNothing );
      xbgiVPrintf( XBGI->dpy, pwin, XBGI->gc, XBGI->fs, textjust, fmt, argptr );
      break;      
    case ButtonPress:
    case KeyPress:
      XUnmapSubwindows( XBGI->dpy, rwin );
      XUnmapWindow( XBGI->dpy, rwin );
      XDestroySubwindows( XBGI->dpy, rwin );
      XDestroyWindow( XBGI->dpy, rwin );
      return;
    }
  }
}
void bgiprintf( char *title, int textjust, char *fmt, ... )
{
  va_list argptr;
  
  va_start( argptr, fmt );
  bgivprintf( title, textjust, fmt, argptr );
}
void bgigets( char *title, int maxlen, char *str )
{
  char *OK = " ok ", *CANCEL = " cancel ";
  Window rwin, twin, swin, owin, cwin;
  XSetWindowAttributes att;
  Cursor curs;
  XEvent ev, exposeev;
  char *bytes = NULL, c;
  KeySym key;
  int nbytes, width, height, textwidth, textheight, textstartx, textstarty;
  int x, y, owidth, cwidth;
  Time clicktime;
  int flg = !0;

  textwidth  = XTextWidth( XBGI->fs, "_", 1 )*20;
  textheight = TextHeight( XBGI->fs );
  width  = Brim( textwidth  );
  height = Brim( textheight );
  textstartx = HemBrim;
  textstarty = HemBrim + TextAscent(XBGI->fs);
  x = XBGI->button_press_ev.xbutton.x_root;
  y = XBGI->button_press_ev.xbutton.y_root;
  x = x - Hem(width)/2;
  y = y - Hem(height)/2;
  owidth = Brim( XTextWidth(XBGI->fs,OK,    strlen(OK)    ) );
  cwidth = Brim( XTextWidth(XBGI->fs,CANCEL,strlen(CANCEL)) );
  rwin = XCreateSimpleWindow( XBGI->dpy, RootWindow( XBGI->dpy, 0 ),
			     x, y, Hem(Hem(width)), Hem(3*Hem(height)), 0, XBGI->fg, XBGI->bg );
  XSelectInput( XBGI->dpy, rwin, KeyPressMask|ButtonPressMask );
  twin = XCreateSimpleWindow( XBGI->dpy, rwin,
			     DecoHemBits, DecoHemBits, Hem(width), Hem(height), 0,
			     XBGI->fg, XBGI->bg );
  swin = XCreateSimpleWindow( XBGI->dpy, rwin,
			     DecoHemBits, DecoHemBits+Hem(height), Hem(width), Hem(height), 0,
			     XBGI->fg, XBGI->bg );
  owin = XCreateSimpleWindow( XBGI->dpy, rwin,
			     DecoHemBits, DecoHemBits+2*Hem(height),
			     Hem(owidth), Hem(height), 0, XBGI->fg, XBGI->bg );
  cwin = XCreateSimpleWindow( XBGI->dpy, rwin,
			     DecoHemBits+Hem(owidth), DecoHemBits+2*Hem(height),
			     Hem(cwidth), Hem(height), 0, XBGI->fg, XBGI->bg );
  XSelectInput( XBGI->dpy, twin, ExposureMask|KeyPressMask|ButtonPressMask );
  XSelectInput( XBGI->dpy, swin, ExposureMask|KeyPressMask|ButtonPressMask );
  XSelectInput( XBGI->dpy, owin, ExposureMask|KeyPressMask|ButtonPressMask );
  XSelectInput( XBGI->dpy, cwin, ExposureMask|KeyPressMask|ButtonPressMask );
  curs = XCreateFontCursor( XBGI->dpy, XC_xterm );
  XDefineCursor( XBGI->dpy, swin, curs );
  XFlush( XBGI->dpy );
  XMapWindow( XBGI->dpy, rwin );
  XMapSubwindows( XBGI->dpy, rwin );
  XSetTransientForHint( XBGI->dpy, rwin, XBGI->root );
  do
    XNextEvent( XBGI->dpy, &ev );
  while ( ev.type != Expose );
  XPutBackEvent( XBGI->dpy, &ev );

  XFlush( XBGI->dpy );
  while ( flg ){
    XNextEvent( XBGI->dpy, &ev );
    switch ( ev.type ){
    case Expose:
      textwidth = XTextWidth( XBGI->fs, str, strlen(str) ) + XTextWidth( XBGI->fs, " ", 1 );
      if ( textwidth < width )
	textstartx = 0;
      else
	textstartx = width - textwidth;
      if ( ev.xexpose.window == rwin ){
	xbgiDecorateWindow( XBGI->dpy, rwin, XBGI->gc,
			   XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
      }else if ( ev.xexpose.window == twin ){
	xbgiDecorateWindow( XBGI->dpy, twin, XBGI->gc,
			   XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
	xbgiPrintf( XBGI->dpy, twin, XBGI->gc, XBGI->fs, CENTER_TEXT, "%s", title );
      }else if ( ev.xexpose.window == swin ){
	exposeev = ev;
	xbgiDecorateWindow( XBGI->dpy, swin, XBGI->gc,
			   XBGI->fg, XBGI->lc, XBGI->dc, DecoRectBack );
	XSetForeground( XBGI->dpy, XBGI->gc, XBGI->fg );
	XSetBackground( XBGI->dpy, XBGI->gc, XBGI->bg );
	XDrawString( XBGI->dpy, swin, XBGI->gc, textstartx, textstarty, str, strlen(str) );
	XSetForeground( XBGI->dpy, XBGI->gc, XBGI->bg );
	XSetBackground( XBGI->dpy, XBGI->gc, XBGI->fg );
	XDrawImageString( XBGI->dpy, swin, XBGI->gc, 
			 textstartx + XTextWidth( XBGI->fs, str, strlen(str) ), textstarty,
			 " ", 1 );
	if ( textstartx < 0 ){
	  XSetForeground( XBGI->dpy, XBGI->gc, XBGI->bg );
	  XSetBackground( XBGI->dpy, XBGI->gc, XBGI->fg );
	  XDrawImageString( XBGI->dpy, swin, XBGI->gc, 0, textstarty, "<", 1 );
	  XSetForeground( XBGI->dpy, XBGI->gc, XBGI->fg );
	  XSetBackground( XBGI->dpy, XBGI->gc, XBGI->bg );
	}
      }else if ( ev.xexpose.window == owin ){
	xbgiDecorateWindow( XBGI->dpy, owin, XBGI->gc,
			   XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
	xbgiPrintf( XBGI->dpy, owin, XBGI->gc, XBGI->fs, CENTER_TEXT, OK );
      }else{
	xbgiDecorateWindow( XBGI->dpy, cwin, XBGI->gc,
			   XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
	xbgiPrintf( XBGI->dpy, cwin, XBGI->gc, XBGI->fs, CENTER_TEXT, CANCEL );
      }
      XFlush( XBGI->dpy );
      break;
    case KeyPress:
      if ( XLookupString( (XKeyEvent*)&ev, &c, 1, &key, NULL ) == 1 ){
	switch ( key ){
	case XK_Escape:
	  strcpy( str, "" );
	  flg = 0;
	  break;
	case XK_Return:
	  flg = 0;
	  break;
	case XK_BackSpace:
	case XK_Delete:
	  str[ max(0,strlen(str)-1) ] = '\0';
	  break;
	case XK_Tab:
	  break;
	case XK_space:
	  break;
	default:
	  if ( !IsCursorKey(key) && !IsFunctionKey(key) && !iscntrl(c) ){
	    str[ min(strlen(str)+1,maxlen-1) ] = '\0';
	    str[ min(strlen(str),  maxlen-2) ] = c;
	  }else if ( iscntrl(c) ){
	    switch(c){
	    case CTRL('w'):
	      strcpy( str, "" );
	      break;
	    }
	  }
	  break;
	}
      }
      XPutBackEvent( XBGI->dpy, &exposeev );
      break;
    case ButtonPress:
      switch ( ev.xbutton.button ){
      case Button1:
	if ( ev.xbutton.window == owin )
	  flg = 0;
	else if ( ev.xbutton.window == cwin ){
	  strcpy( str, "" );
	  flg = 0;
	}
	break;
      case Button2:
        bytes = XFetchBytes( XBGI->dpy, &nbytes );
	bytes[nbytes] = '\0';
	if ( strlen(str)+nbytes < maxlen ){
	  strcat( str, bytes );
	}
	XPutBackEvent( XBGI->dpy, &exposeev );
        break;
      }
    }
  }
  XUnmapSubwindows( XBGI->dpy, rwin );
  XUnmapWindow( XBGI->dpy, rwin );
}
void bgimenu( char *title, int menuc, char *menuv[], int *menui )
{
  Window rwin, twin, *mwin;
  XSetWindowAttributes att;
  XEvent ev;
  int i, width, height, textwidth, textheight, x, y;
  Window root, child;
  int rx, ry, wx, wy;
  unsigned int mask;
  
  xbgiGetSizePrintf( XBGI->fs, &textwidth, &textheight, "%s", title );
  for ( i=0; i<menuc; i++ ){
    xbgiGetSizePrintf( XBGI->fs, &width, &textheight, "%s", menuv[i] );
    if ( textwidth < width )
      textwidth = width;
  }
  width  = textwidth;
  height = textheight;
  x = XBGI->button_press_ev.xbutton.x_root;
  y = XBGI->button_press_ev.xbutton.y_root;
  x = x - Hem(width)/2;
  y = y - Hem(height)/2;
  rwin = XCreateSimpleWindow( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
			     x, y, Hem(Hem(width)), Hem(Hem(height)*(menuc+1)), 0,
			     XBGI->fg, XBGI->bg );
  XSelectInput( XBGI->dpy, rwin, ExposureMask|Button2MotionMask );

  att.override_redirect = True;
  XChangeWindowAttributes( XBGI->dpy, rwin, CWOverrideRedirect, &att );
  twin = XCreateSimpleWindow( XBGI->dpy, rwin,
			     DecoHemBits, DecoHemBits, Hem(width), Hem(height), 0,
			     XBGI->fg, XBGI->bg );
  mwin = (Window*)malloc( (menuc)*sizeof(Window) );
  for ( i=0; i<menuc; i++ ){
    mwin[i] = XCreateSimpleWindow( XBGI->dpy, rwin, 
				  DecoHemBits, DecoHemBits+Hem(height)*(i+1),
				  Hem(width), Hem(height), 0, XBGI->fg, XBGI->bg );
  }
  XFlush( XBGI->dpy );
  XMapWindow( XBGI->dpy, rwin );
  XMapSubwindows( XBGI->dpy, rwin );
  do{
    XNextEvent( XBGI->dpy, &ev );
  }while ( Expose != ev.type );
  XPutBackEvent( XBGI->dpy, &ev );

  XFlush( XBGI->dpy );
  while (1){
    XNextEvent( XBGI->dpy, &ev );
    switch ( ev.type ){
    case Expose:
      xbgiDecorateWindow( XBGI->dpy, rwin, XBGI->gc,
			 XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
      xbgiDecorateWindow( XBGI->dpy, twin, XBGI->gc,
			 XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
      xbgiPrintf( XBGI->dpy, twin, XBGI->gc, XBGI->fs, CENTER_TEXT, "%s", title );
      for ( i=0; i<menuc; i++ ){
	xbgiDecorateWindow( XBGI->dpy, mwin[i], XBGI->gc,
			   XBGI->fg, XBGI->lc, XBGI->dc, DecoNothing );
	xbgiPrintf( XBGI->dpy, mwin[i], XBGI->gc, XBGI->fs, LEFT_TEXT, "%s", menuv[i] );
      }
      XFlush( XBGI->dpy );
      break;
    /***case EnterNotify:
      for ( i=0; i<menuc; i++ ){
	if ( ev.xcrossing.window == mwin[i] ){
	  xbgiDecorateWindow( XBGI->dpy, mwin[i], XBGI->gc, 
			     XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
	  xbgiPrintf( XBGI->dpy, mwin[i], XBGI->gc, XBGI->fs, LEFT_TEXT, "%s", menuv[i] );
	}
      }
      XFlush( XBGI->dpy );
      break;
    case LeaveNotify:
      for ( i=0; i<menuc; i++ ){
	if ( ev.xcrossing.window == mwin[i] ){
	  xbgiDecorateWindow( XBGI->dpy, mwin[i], XBGI->gc,
			     XBGI->fg, XBGI->lc, XBGI->dc, DecoNothing );
	  xbgiPrintf( XBGI->dpy, mwin[i], XBGI->gc, XBGI->fs, LEFT_TEXT, "%s", menuv[i] );
	}
      }
      XFlush( XBGI->dpy );
      break;***/
    case MotionNotify:
      XQueryPointer( XBGI->dpy, rwin, &root, &child, &rx, &ry, &wx, &wy, &mask );
      if (( child == 0 )||( child == twin )){
	if ( *menui != -1 ){
	  xbgiDecorateWindow( XBGI->dpy, mwin[*menui], XBGI->gc,
			     XBGI->fg, XBGI->lc, XBGI->dc, DecoNothing );
	  xbgiPrintf( XBGI->dpy, mwin[*menui], XBGI->gc, XBGI->fs, 
		     LEFT_TEXT, "%s", menuv[*menui] );
	}
	*menui = -1;
      }else{
	for ( i=0; i<menuc; i++ )
	  if ( child == mwin[i] ){
	    if (( *menui != -1 )&&( *menui != i )){
	      xbgiDecorateWindow( XBGI->dpy, mwin[*menui], XBGI->gc,
				 XBGI->fg, XBGI->lc, XBGI->dc, DecoNothing );
	      xbgiPrintf( XBGI->dpy, mwin[*menui], XBGI->gc, XBGI->fs, 
			 LEFT_TEXT, "%s", menuv[*menui] );
	    }
	    xbgiDecorateWindow( XBGI->dpy, mwin[i], XBGI->gc,
			       XBGI->fg, XBGI->lc, XBGI->dc, DecoRectFront );
	    xbgiPrintf( XBGI->dpy, mwin[i], XBGI->gc, XBGI->fs, 
		       LEFT_TEXT, "%s", menuv[i] );
	    *menui = i;
	    break;
	  }
      }
      break;
    case ButtonRelease:
      XUnmapSubwindows( XBGI->dpy, rwin );
      XUnmapWindow( XBGI->dpy, rwin );
      XDestroySubwindows( XBGI->dpy, rwin );
      free( mwin );
      XDestroyWindow( XBGI->dpy, rwin );
      return;
      break;
    }
  }
}
