/*
 * Source File < graphics.c >
 *
 * Standard Functions: Borland Graphics Interface for Xlib
 *
 * Copyleft (c) 1993 Tokyo Denki University, Taiji Yamada
 */
#include <malloc.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include "option.h"
#include "graphix.h"
#include "graphics.h"
#include "xbgiutl.h"
#include "bgiutl.h"
#include "_bgi.h"

#define ENABLE_XPMLIB

/*
 * the header of XBGI to stack and current XBGI for graphics
 */
xbgi *XBGIS = NULL;
xbgi *XBGI  = NULL;

/*
 * internal static functions
 */
static void rferror( char* err )
{
  static int flg = !0;

  if ( flg )
    fprintf( stderr, "XBGI: Reserved Future.\n -> %s \n", err );
  flg = 0;
}
static void nserror( char* err )
{
  static int flg = !0;

  if ( flg )
    fprintf( stderr, "XBGI: Not Supported.  \n -> %s \n", err );
  flg = 0;
}
static void polypoints2xpoints( int n, int polypoints[], XPoint *xpoints )
{
  int i;

  for ( i=0; i<n; i++ ){
    xpoints->x = Xx( *polypoints++ );
    xpoints->y = Xy( *polypoints++ );
    xpoints++;
  }
}
static unsigned long XGetPixelColorName( Display *d, char *colorname )
{
  Colormap cmap;
  XColor c0, c1;

  cmap = DefaultColormap( d, 0 );
  XAllocNamedColor( d, cmap, colorname, &c1, &c0 );
  return( c1.pixel );
}

/*
 * emulation function in DOS.H
 */
void delay( unsigned long ms )
{
  unsigned long us = ms*1000;
  struct timeval tval;

  tval.tv_sec  = us/1000000;
  tval.tv_usec = us%1000000;
  select( 0, NULL, NULL, NULL, &tval );
}

/*
 * official enhanced functions for X interface
 */
void xflush( void )
{
  XFlush( XBGI->dpy );
}
void settitle( char *title )
{
  XStoreName( XBGI->dpy, XBGI->root, title );
  XSetIconName( XBGI->dpy, XBGI->root, title );
}
void seticonname( char *name )
{
  XSetIconName( XBGI->dpy, XBGI->root, name );
}
void setmaxx( int maxx )
{
  Window root;
  int i, x, y, width, height, border, depth;
  
  if ( maxx <= 0 )
    return;
  XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
               &x, &y, &width, &height, &border, &depth );
  XResizeWindow( XBGI->dpy, XBGI->win[XBGI->apage], maxx+1, height );
  if ( XBGI->apage == 0 ){
    for ( i=0; i<XBGI->npage; i++ )
      XResizeWindow( XBGI->dpy, XBGI->win[i], maxx+1, height );
    XResizeWindow( XBGI->dpy, XBGI->root, maxx+1, height );
  }
}
void setmaxy( int maxy )
{
  Window root;
  int i, x, y, width, height, border, depth;

  if ( maxy <= 0 )
    return;
  XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
               &x, &y, &width, &height, &border, &depth );
  XResizeWindow( XBGI->dpy, XBGI->win[XBGI->apage], width, maxy+1 );
  if ( XBGI->apage == 0 ){
    for ( i=0; i<XBGI->npage; i++ )
      XResizeWindow( XBGI->dpy, XBGI->win[i], width, maxy+1 );
    XResizeWindow( XBGI->dpy, XBGI->root, width, maxy+1 );
  }
}
void setmaxcolor( int c )
{
  int i, cc;
  unsigned long plane_mask;
  struct rgbtype rgb[8] = {
    { 0, 0, 0 },
    { 0, 0, 1 },
    { 0, 1, 0 },
    { 0, 1, 1 },
    { 1, 0, 0 },
    { 1, 0, 1 },
    { 1, 1, 0 },
    { 1, 1, 1 },
  };
  
  switch( XBGI->gmode ){
  case BGIMONO:
    XBGI->palette.size = 2;
    XBGI->colortable[BlackPixel(XBGI->dpy,XBGI->scr)] = BLACK;
    XBGI->colortable[WhitePixel(XBGI->dpy,XBGI->scr)] = WHITE;
    XBGI->pixels[BLACK] = BlackPixel( XBGI->dpy, XBGI->scr );
    XBGI->pixels[WHITE] = WhitePixel( XBGI->dpy, XBGI->scr );
    for ( i=0; i<XBGI->palette.size; i++ )
      XBGI->palette.colors[i] = i;
    break;
  case BGIGRAY:
    XBGI->palette.size = c + 1;
    XAllocColorCells( XBGI->dpy, XBGI->colormap,
                     False, &plane_mask, 0, XBGI->pixels, XBGI->palette.size );
    for ( i=0; i<XBGI->palette.size; i++ ){
      XBGI->colortable[XBGI->pixels[i]] = i;
      XBGI->palette.colors[i] = i;
    }
    for ( i=1; i<XBGI->palette.size; i++ )
      setrgbpalette( i, i, i, i );
    setallpalette( &XBGI->palette );
    break;
  case BGICOLOR:
    if ( XBGI->gerror != grNoInitGraph )
      XFreeColors( XBGI->dpy, XBGI->colormap, XBGI->pixels, XBGI->palette.size, 0 );
    XBGI->palette.size = c + 1;
    XAllocColorCells( XBGI->dpy, XBGI->colormap,
                     False, &plane_mask, 0, XBGI->pixels, XBGI->palette.size );
    for ( i=0; i<XBGI->palette.size; i++ ){
      XBGI->colortable[XBGI->pixels[i]] = i;
      XBGI->palette.colors[i] = i;
    }
    for ( i=0; i<XBGI->palette.size/8; i++ )
      for ( cc=0; cc<8; cc++ )
	setrgbpalette( 8*i+cc, 8*(i+1)*rgb[cc].r, 8*(i+1)*rgb[cc].g, 8*(i+1)*rgb[cc].b );
    for ( i=1; i<XBGI->palette.size/8; i++ )
      setrgbpalette( 8*i+0, 4*(i+1), 4*(i+1), 4*(i+1) );
    cc = XBGI->palette.size/2;
    for ( i=0; i<XBGI->palette.size/8; i++ )
      setrgbpalette( 8*i+7, cc+4*(i+1), cc+4*(i+1), cc+4*(i+1) );
    /*setallpalette( &XBGI->palette );*/
    break;
  }
}
void setbackingstore( int flg )
{
  int i;
  XSetWindowAttributes att;

  if ( flg )
    att.backing_store = Always;
  else
    att.backing_store = NotUseful;
  XChangeWindowAttributes( XBGI->dpy, XBGI->win[XBGI->apage], CWBackingStore, &att );
}
int graphnumber( void )
{
  return( XBGI->id );
}
void selectgraph( int id )
{
  xbgi *ptr = XBGIS;

  while ( ptr != NULL ){
    if ( ptr->id == id ){
      XBGI = ptr;
      return;
    }
    ptr = ptr->next;
  }
}
void termgraph( int id )
{
  xbgi *ptr = XBGIS;

  while ( ptr != NULL ){
    if ( ptr->id == id ){
      destroygraph( ptr );
      return;
    }
    ptr = ptr->next;
  }
}

/*
 *  functions for draw line
 */
int getx( void )
{
  return( XBGI->cp.x );
}
int gety( void )
{
  return( XBGI->cp.y );
}
void getlinesettings( struct linesettingstype *lineinfo )
{
  *lineinfo = XBGI->lineset;
}
/*static void showbit( unsigned pattern )
{
  int i;

  for ( i=0; i<16; i++ )
    if ( pattern&(0x8000>>i) )
      printf( "1" );
    else
      printf( "0" );
}*/
static void upattern2xlinestyle( unsigned upattern, int *offset, char *usrbit, int *n )
{
  int i, j, c, ns, flg;
  unsigned long ul;

  /*showbit( upattern );
  printf( "\n" );*/
  ul = (unsigned long)upattern*0x10000 + upattern;
  ns = 0;
  while ( (!(ul&(0x80000000>>ns)))&&(ns<32) ) ns++;
  if ( ns == 32 ){
    *n = 0;
    return;
  }
  upattern = (ul<<ns)>>16;
  /*showbit( upattern );
  printf( "\n" );*/
  *offset = i = c = 0;
  flg = !0;
  while ( i < 16-ns ){
    if ( flg ){
      if ( !(upattern&(0x8000>>i)) ){
        c++;
        flg = !flg;
      }
      usrbit[c]++;
    }else{
      if ( upattern&(0x8000>>i) ){
        c++;
        flg = !flg;
      }
      usrbit[c]++;
    }
    i++;
  }
  c++;
  if ( ns > 0 ){
    if ( !flg ){
      usrbit[c] = 1; /* ? */
      c++;
    }
    usrbit[c] = ns;
    c++;
    *offset = c-1;
  }
  *n = c;
  /*printf( "offset(%d)", *offset );
  for ( i=0; i<c; i++ )
    printf( " %u", usrbit[i] );
  printf( "num(%d)\n", *n );*/
}
void setlinestyle( int linestyle, unsigned upattern, int thickness )
{
  static char dotted[] = { 1, 3 };
  static char center[] = { 1, 3, 4, 3 };
  static char dashed[] = { 7, 4 };
  char usrbit[] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  int offset = 0, n = 0;

  XBGI->lineset.linestyle = linestyle;
  XBGI->lineset.upattern  = upattern;
  XBGI->lineset.thickness = thickness;
  switch( linestyle ){
  case SOLID_LINE:
    break;
  case DOTTED_LINE:
    XSetDashes( XBGI->dpy, XBGI->gcdraw, 0, dotted, 2 );
    break;
  case CENTER_LINE:
    XSetDashes( XBGI->dpy, XBGI->gcdraw, 0, center, 4 );
    break;
  case DASHED_LINE:
    XSetDashes( XBGI->dpy, XBGI->gcdraw, 0, dashed, 2 );
    break;
  case USERBIT_LINE:
    upattern2xlinestyle( upattern, &offset, usrbit, &n );
    if ( n != 0 )
      XSetDashes( XBGI->dpy, XBGI->gcdraw, offset, usrbit, n );
    else
      nserror( "setlinestyle( USERBIT_LINE, 0, )" );
    break;
  }
  XSetLineAttributes( XBGI->dpy, XBGI->gcdraw, thickness,
                     ((linestyle==SOLID_LINE)?LineSolid:LineOnOffDash), CapButt, JoinBevel );
}
void line( int x1, int y1, int x2, int y2 )
{
  XDrawLine( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
            Xx( x1 ), Xy( y1 ), Xx( x2 ), Xy( y2 ) );
}
void linerel( int dx, int dy )
{
  XDrawLine( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
            Xx( XBGI->cp.x ), Xy( XBGI->cp.y ),
            Xx( XBGI->cp.x + dx ), Xy( XBGI->cp.y + dy ) );
  XBGI->cp.x += dx;
  XBGI->cp.y += dy;
}
void lineto( int x, int y )
{
  XDrawLine( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
            Xx( XBGI->cp.x ), Xy( XBGI->cp.y ), Xx( x ), Xy( y ) );
  XBGI->cp.x = x;
  XBGI->cp.y = y;
}
void moverel( int dx, int dy )
{
  XBGI->cp.x += dx;
  XBGI->cp.y += dy;
}
void moveto( int x, int y )
{
  XBGI->cp.x = x;
  XBGI->cp.y = y;
}

/*
 * functions for fillstyle
 */
#include "bitmaps/empty_fill.xbm"
#include "bitmaps/solid_fill.xbm"
#include "bitmaps/line_fill.xbm"
#include "bitmaps/ltslash_fill.xbm"
#include "bitmaps/slash_fill.xbm"
#include "bitmaps/bkslash_fill.xbm"
#include "bitmaps/ltbkslash_fill.xbm"
#include "bitmaps/hatch_fill.xbm"
#include "bitmaps/xhatch_fill.xbm"
#include "bitmaps/interleave_fill.xbm"
#include "bitmaps/wide_dot_fill.xbm"
#include "bitmaps/close_dot_fill.xbm"
#include "bitmaps/user_fill.xbm"
static char *fill_bits[USER_FILL+1] = {
  empty_fill_bits,
  solid_fill_bits,
  line_fill_bits,
  ltslash_fill_bits,
  slash_fill_bits,
  bkslash_fill_bits,
  ltbkslash_fill_bits,
  hatch_fill_bits,
  xhatch_fill_bits,
  interleave_fill_bits,
  wide_dot_fill_bits,
  close_dot_fill_bits,
  user_fill_bits,
};
void getfillpattern( char *pattern )
{
  int i;

  for ( i=0; i<8; i++ )
    pattern[i] = fill_bits[XBGI->fillset.pattern][i];
}
void getfillsettings( struct fillsettingstype *fillinfo )
{
  *fillinfo = XBGI->fillset;
}
void setfillpattern( char *upattern, int color )
{
  int i;
  Pixmap pm_fill;

  XBGI->fillset.pattern = USER_FILL;
  XBGI->fillset.color   = minmax( 0, color, XBGI->palette.size-1 );
  for ( i=0; i<8; i++ )
    user_fill_bits[i] = upattern[i];
  XSetForeground( XBGI->dpy, XBGI->gcfill, XBGI->pixels[XBGI->fillset.color] );
  pm_fill = XCreateBitmapFromData( XBGI->dpy, XBGI->root, user_fill_bits, 8, 8 );
  XSetStipple( XBGI->dpy, XBGI->gcfill, pm_fill );
  XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillOpaqueStippled );
}
void setfillstyle( int pattern, int color )
{
  Pixmap pm_fill;

  XBGI->fillset.pattern = minmax( 0, pattern, USER_FILL );
  XBGI->fillset.color   = minmax( 0, color, XBGI->palette.size-1 );
  XSetForeground( XBGI->dpy, XBGI->gcfill, XBGI->pixels[XBGI->fillset.color] );
  switch( XBGI->fillset.pattern ){
  case EMPTY_FILL:
    pm_fill = XCreateBitmapFromData( XBGI->dpy, XBGI->root,
                                    fill_bits[XBGI->fillset.pattern], 8, 8 );
    XSetStipple( XBGI->dpy, XBGI->gcfill, pm_fill );
    XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillStippled );
    break;
  case SOLID_FILL:
    XSetForeground( XBGI->dpy, XBGI->gcfill, XBGI->pixels[XBGI->fillset.color] );
    XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillSolid );
    break;
  case LINE_FILL:
  case LTSLASH_FILL:
  case SLASH_FILL:
  case BKSLASH_FILL:
  case LTBKSLASH_FILL:
  case HATCH_FILL:
  case XHATCH_FILL:
  case INTERLEAVE_FILL:
  case WIDE_DOT_FILL:
  case CLOSE_DOT_FILL:
  case USER_FILL:
    pm_fill = XCreateBitmapFromData( XBGI->dpy, XBGI->root,
                                    fill_bits[XBGI->fillset.pattern], 8, 8 );
    XSetStipple( XBGI->dpy, XBGI->gcfill, pm_fill );
    XSetFillStyle( XBGI->dpy, XBGI->gcfill, FillOpaqueStippled );
    break;
  }
}
void floodfill( int x, int y, int border )
{
  int s[1000][2];
  int p, cx, cy, u, u2, d, d2, pu, pd;
  int minx, miny, maxx, maxy;

  if ( XBGI->viewport.clip ){
    minx = XBGI->viewport.left;
    miny = XBGI->viewport.top;
    maxx = XBGI->viewport.right;
    maxy = XBGI->viewport.bottom;
  }else{
    minx = 0;
    miny = 0;
    maxx = getmaxx();
    maxy = getmaxy();
  }
  p = 1;
  s[p][0] = x;
  s[p][1] = y;
  while ( p > 0 ){
    cx = s[p][0];
    cy = s[p--][1];
    if ( getpixel( cx, cy ) != border ){
      while ( getpixel( cx, cy )!=border && cx>=minx ) cx--;
      cx++;
      u = 0;
      d = 0;
      while ( getpixel( cx, cy )!=border && cx<=maxx ){
        putpixel( cx, cy, XBGI->fillset.color );
        u2 = u;
        if ( cy > miny ){
          pu = getpixel( cx, cy-1 );
          if ( pu == border || pu == XBGI->fillset.color )
	    u = 0;
	  else
	    u = 1;
        }
        if ( u2==0 && u==1 ){
	  p++;
	  s[p][0] = cx;
	  s[p][1] = cy - 1;
	}
        d2 = d;
        if ( cy < maxy ) {
          pd = getpixel( cx, cy+1 );
          if ( pd == border || pd == XBGI->fillset.color )
	    d = 0;
          else
	    d = 1;
        }
        if ( d2==0 && d==1 ){
	  p++;
	  s[p][0] = cx;
	  s[p][1] = cy + 1;
	}
        cx++;
      }
    }
  }
}

/*
 * functions for draw/fill rectangle
 */
void rectangle( int left, int top, int right, int bottom )
{
  XDrawRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left), abs(bottom-top) );
}
void bar( int left, int top, int right, int bottom )
{
  XFillRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left)+1, abs(bottom-top)+1 );
  /*XDrawRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                   min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                   abs(right-left), abs(bottom-top) );*/
}
void bar3d( int left, int top, int right, int bottom, int depth, int topflag )
{
  int topheight = 0;
  XPoint xpoints[4];

  if ( topflag )
    topheight = depth/2;
  xpoints[0].x = Xx( right );         xpoints[0].y = Xy( top );
  xpoints[1].x = Xx( right + depth ); xpoints[1].y = Xy( top - topheight );
  xpoints[2].x = Xx( left + depth );  xpoints[2].y = Xy( top - topheight );
  xpoints[3].x = Xx( left );          xpoints[3].y = Xy( top );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw, xpoints, 4, CoordModeOrigin );
  xpoints[0].x = Xx( right );         xpoints[0].y = Xy( top );
  xpoints[1].x = Xx( right );         xpoints[1].y = Xy( bottom );
  xpoints[2].x = Xx( right + depth ); xpoints[2].y = Xy( bottom - topheight );
  xpoints[3].x = Xx( right + depth ); xpoints[3].y = Xy( top - topheight );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw, xpoints, 4, CoordModeOrigin );
  XFillRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left)+1, abs(bottom-top)+1 );
  XDrawRectangle( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                 min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                 abs(right-left), abs(bottom-top) );
}

/*
 * functions for draw/fill poly
 */
void drawpoly( int numpoints, int *polypoints )
{
  XPoint *xpoints = (XPoint *)malloc( numpoints * sizeof(XPoint) );
  
  polypoints2xpoints( numpoints, polypoints, xpoints );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
             xpoints, numpoints, CoordModeOrigin );
  free( xpoints );
}
void fillpoly( int numpoints, int *polypoints )
{
  XPoint *xpoints = (XPoint *)malloc( numpoints * sizeof(XPoint) );
  
  polypoints2xpoints( numpoints, polypoints, xpoints );
  XFillPolygon( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
               xpoints, numpoints, Complex, CoordModeOrigin );
  XDrawLines( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
             xpoints, numpoints, CoordModeOrigin );
  free( xpoints );
}

/*
 * functions for draw/fill arc
 */
struct arcsettingstype {
  int x, y;
  int stangle, endangle;
  int xradius, yradius;
};
static struct arcsettingstype arcset = { 0, 0, 0, 360, 100, 100 };

void getaspectratio( int *xasp, int *yasp )
{
  *xasp = 10000; *yasp = 10000;
}
void setaspectratio( int xasp, int yasp )
{
  rferror( "setaspectratio()" );
}
void getarccoords( struct arccoordstype *arccoords )
{
  double stangle, endangle;
  
  stangle  = (double)arcset.stangle* M_PI/180;
  endangle = (double)arcset.endangle*M_PI/180;
  arccoords->x      = arcset.x;
  arccoords->y      = arcset.y;
  arccoords->xstart = arcset.x + (int)((double)arcset.xradius*cos(stangle ));
  arccoords->ystart = arcset.y - (int)((double)arcset.yradius*sin(stangle ));
  arccoords->xend   = arcset.x + (int)((double)arcset.xradius*cos(endangle));
  arccoords->yend   = arcset.y - (int)((double)arcset.yradius*sin(endangle));
}
void arc( int x, int y, int stangle, int endangle, int radius )
{
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = radius;
  arcset.yradius  = radius;
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius,
           64 * stangle, 64 * ( endangle - stangle ) );
}
void pieslice( int x, int y, int stangle, int endangle, int radius )
{
  static struct arccoordstype arccoords;
  
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = radius;
  arcset.yradius  = radius;
  getarccoords( &arccoords );
  XFillArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius,
           64 * stangle, 64 * ( endangle - stangle ) );
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius,
           64 * stangle, 64 * ( endangle - stangle ) );
  line( arccoords.xstart, arccoords.ystart, arccoords.x, arccoords.y );
  line( arccoords.xend,   arccoords.yend,   arccoords.x, arccoords.y );
}
void circle( int x, int y, int radius )
{
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = 0;
  arcset.endangle = 360;
  arcset.xradius  = radius;
  arcset.yradius  = radius;
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - radius ), Xy( y - radius ), 2 * radius, 2 * radius, 0, 64 * ( 360 - 0 ) );
}
void ellipse( int x, int y, int stangle, int endangle, int xradius, int yradius )
{
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = xradius;
  arcset.yradius  = yradius;
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - xradius ), Xy( y - yradius ), 2 * xradius, 2 * yradius,
           64 * stangle, 64 * ( endangle - stangle ) );
}
void sector( int x, int y, int stangle, int endangle, int xradius, int yradius )
{
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = stangle;
  arcset.endangle = endangle;
  arcset.xradius  = xradius;
  arcset.yradius  = yradius;
  XFillArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
           Xx( x - xradius ), Xy( y - yradius ), 2 * xradius, 2 * yradius,
           64 * stangle, 64 * ( endangle - stangle ) );
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - xradius ), Xy( y - yradius ), 2 * xradius, 2 * yradius,
           64 * stangle, 64 * ( endangle - stangle ) );
}
void fillellipse( int x, int y, int xradius, int yradius )
{
  arcset.x        = x;
  arcset.y        = y;
  arcset.stangle  = 0;
  arcset.endangle = 360;
  arcset.xradius  = xradius;
  arcset.yradius  = yradius;
  XFillArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcfill,
           Xx( x - xradius ), Xy( y - yradius ) , 2 * xradius, 2 * yradius, 0, 64 * ( 360 - 0 ) );
  XDrawArc( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
           Xx( x - xradius ), Xy( y - yradius ) , 2 * xradius, 2 * yradius, 0, 64 * ( 360 - 0 ) );
}

/*
 * functions for clear graphics
 */
void cleardevice( void )
{
  XClearWindow( XBGI->dpy, XBGI->win[XBGI->apage] );
}
void clearviewport( void )
{
  if ( XBGI->viewport.clip )
    XClearArea( XBGI->dpy, XBGI->win[XBGI->apage],
               XBGI->viewport.left, XBGI->viewport.top, 
               XBGI->viewport.right - XBGI->viewport.left+1,
               XBGI->viewport.bottom - XBGI->viewport.top+1, False );
  else
    cleardevice();
}

/*
 * functions for standard creation and termination graphics
 */
void detectgraph( int *graphdriver, int *graphmode )
{
  XVisualInfo vinfo_ret;
  Status result;
  
  /*
   * case of detecting before initgraph, always BGIC16
   */
  if ( XBGIS == NULL ){
    *graphdriver = BGI;
    *graphmode   = BGICOLOR;
    return;
  }
  /*
   * case of detecting after initgraph, research visual class and depth
   */
  switch( *graphdriver ){
  case DETECT:
    *graphdriver = XBGI->gdriver = BGI;
    *graphmode   = XBGI->gmode   = BGICOLOR;
    break;
  case BGI:
  case BGIS:
    break;
  case BGITGIF:
    rferror( "initgraph( BGITGIF, )" );
  case BGIPS:
    rferror( "initgraph( BGIPS, )" );
  default:
    *graphdriver = XBGI->gdriver = BGI;
    *graphmode   = XBGI->gmode   = BGICOLOR;
    break;
  }
  XBGI->depth  = DefaultDepth( XBGI->dpy, XBGI->scr );
  XBGI->vis    = DefaultVisual( XBGI->dpy, XBGI->scr );
  XBGI->vclass = XBGI->vis->class;
  switch ( XBGI->vclass ){
  case StaticGray:
  case GrayScale:
    *graphmode   = XBGI->gmode   = BGIGRAY;
    if ( XBGI->depth == 1 )
      XBGI->gmode = BGIMONO;
    break;
  case StaticColor:
  case TrueColor:    
  case PseudoColor:
  case DirectColor:
    break;
  }
}
void initgraph( int *graphdriver, int *graphmode, char *pathtodriver )
{
  XBGI = creategraph( graphdriver, graphmode, pathtodriver );
}
int graphresult( void )
{
  return( XBGI->gerror );
}
void graphdefaults( void )
{
  int i;
  
  for ( i=0; i<XBGI->npage; i++ ){
    setactivepage( i );
    setviewport( 0, 0, getmaxx(), getmaxy(), 0 ); 
    setcolor( XBGI->color );
    setbkcolor( XBGI->bkcolor );
    cleardevice();
  }
  setvisualpage( 0 );
  setactivepage( 0 );
  setlinestyle( SOLID_LINE, 0, NORM_WIDTH );
  settextstyle( DEFAULT_FONT, HORIZ_DIR, 7 );
  settextjustify( LEFT_TEXT, TOP_TEXT );
  setfillstyle( SOLID_FILL, WHITE );
  moveto( 0, 0 );
}
char* grapherrormsg( int errorcode )
{
  rferror( "grapherrormsg()" );
  return( NULL );
}
void closegraph( void )
{
  destroygraph( XBGI );
}

/*
 * functions for set write function
 */
static int OP2GX[] = {
  GXcopy, GXxor, GXor, GXand, GXcopyInverted,
};
void setwritemode( int op )
{
  XSetFunction( XBGI->dpy, XBGI->gcdraw,  OP2GX[op] );
  XSetFunction( XBGI->dpy, XBGI->gcpixel, OP2GX[op] );
}

/*
 * functions for pixel
 */
unsigned getpixel( int x, int y )
{
  XImage *image;
  int val;

  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    Xx( x ), Xy( y ), 1, 1, AllPlanes, ZPixmap );
  val = XBGI->colortable[image->data[image->xoffset]];
  XDestroyImage( image );
  return( ( unsigned )val );
}
void putpixel( int x, int y, int color )
{
  XSetForeground( XBGI->dpy, XBGI->gcpixel, XBGI->pixels[minmax(0,color,XBGI->palette.size-1)] );
  XDrawPoint( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcpixel, Xx(x), Xy(y) );
}

/*
 * functions for image
 */
unsigned imagesize( int left, int top, int right, int bottom )
{
  return( (unsigned)(2 + (abs(right-left)+1)*(abs(bottom-top)+1)) );
}
void getimage( int left, int top, int right, int bottom, void *bitmap )
{
  XImage *image;
  int width, height, i, j, k;
  unsigned char *data, *buffer = (unsigned char *)bitmap;

  width  = abs(right-left)+1;
  height = abs(bottom-top)+1;
  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    min(Xx(left),Xx(right)), min(Xy(top),Xy(bottom)),
                    width, height, AllPlanes, ZPixmap );
  data = (unsigned char *)image->data;
  *buffer++ = width;
  *buffer++ = height;
  for ( i=0; i<height; i++ ){
    for ( k=j=0; j<image->bytes_per_line; j++, data++ ){
      if ( j<image->xoffset ) continue;
      k++;
      if ( width < k ) continue;
      *buffer++ = XBGI->colortable[*data];
    }
  }
  XDestroyImage( image );
}
void putimage( int left, int top, void *bitmap, int op )
{
  int i;
  int width, height;
  XImage *image;
  unsigned char *src, *dst, *buffer = (unsigned char *)bitmap;
  
  width  = buffer[0];
  height = buffer[1];
  src = buffer + 2;
  dst = (unsigned char *)malloc( width*height*sizeof(unsigned char) );
  for ( i=0; i<width*height; i++ )
    dst[i] = XBGI->pixels[src[i]];
  XSetFunction( XBGI->dpy, XBGI->gcimage, OP2GX[minmax(COPY_PUT,op,NOT_PUT)] );
  image = XCreateImage( XBGI->dpy, DefaultVisual(XBGI->dpy,XBGI->scr), XBGI->depth,
                       ZPixmap, 0, (char *)dst, width, height, 8, 0 );
  XPutImage( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcimage, image,
            0, 0, Xx(left), Xy(top), width, height );
  XDestroyImage( image );
  free( dst );
}

/*
 * functions for page
 */
int getmaxx( void )
{
  Window root;
  int x, y, width, height, border, depth;
  
  XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
               &x, &y, &width, &height, &border, &depth );
  return( width-1 );
}
int getmaxy( void )
{
  Window root;
  int x, y, width, height, border, depth;
  
  XGetGeometry( XBGI->dpy, XBGI->win[XBGI->apage], &root,
               &x, &y, &width, &height, &border, &depth );
  return( height-1 );
}
void setactivepage( int page )
{
  XBGI->apage = minmax( 0, page, XBGI->npage-1 );
}
void setvisualpage( int page )
{
  int i;

  if ( page >= XBGI->npage ){
    /* all pages full mapped gmode */
    XUnmapWindow( XBGI->dpy, XBGI->root );
    for ( i=0; i<XBGI->npage; i++ ){
      XReparentWindow( XBGI->dpy,
                      XBGI->win[i], RootWindow(XBGI->dpy,XBGI->scr), 0, 0 );
      XMapWindow( XBGI->dpy, XBGI->win[i] );
    }
    XBGI->vpage = XBGI->npage;
  }else if ( ( XBGI->vpage >= XBGI->npage )&&( page < XBGI->npage ) ){
    /* one page mapped gmode */
    XUnmapSubwindows( XBGI->dpy, XBGI->root );
    for ( i=0; i<XBGI->npage; i++ )
      XReparentWindow( XBGI->dpy, XBGI->win[i], XBGI->root, 0, 0 );
    XBGI->vpage = minmax( 0, page, XBGI->npage-1 );
    XMapWindow( XBGI->dpy, XBGI->root );
    XMapRaised( XBGI->dpy, XBGI->win[XBGI->vpage] );
  }else{
    XBGI->vpage = minmax( 0, page, XBGI->npage-1 );
    XMapRaised( XBGI->dpy, XBGI->win[XBGI->vpage] );
  }
}

/*
 * functions for viewport
 */
void getviewsettings( struct viewporttype *viewport )
{
  *viewport = XBGI->viewport;
}
void setviewport( int left, int top, int right, int bottom, int clip )
{
  XRectangle rec;

  XBGI->viewport.left   = min( left, right );
  XBGI->viewport.top    = min( top, bottom );
  XBGI->viewport.right  = max( left, right );
  XBGI->viewport.bottom = max( top, bottom );
  XBGI->viewport.clip   = clip;
  if ( XBGI->viewport.clip ){
    rec.x = XBGI->viewport.left;
    rec.y = XBGI->viewport.top;
    rec.width  = XBGI->viewport.right - XBGI->viewport.left + 1;
    rec.height = XBGI->viewport.bottom - XBGI->viewport.top + 1;
    XSetClipRectangles( XBGI->dpy, XBGI->gcdraw,  0, 0, &rec, 1, Unsorted );
    XSetClipRectangles( XBGI->dpy, XBGI->gcfill,  0, 0, &rec, 1, Unsorted );
    XSetClipRectangles( XBGI->dpy, XBGI->gcpixel, 0, 0, &rec, 1, Unsorted );
    XSetClipRectangles( XBGI->dpy, XBGI->gcimage, 0, 0, &rec, 1, Unsorted );
  }else{
    XSetClipMask( XBGI->dpy, XBGI->gcdraw,  None );
    XSetClipMask( XBGI->dpy, XBGI->gcfill,  None );
    XSetClipMask( XBGI->dpy, XBGI->gcpixel, None );
    XSetClipMask( XBGI->dpy, XBGI->gcimage, None );
  }
}

/*
 * function for graph driver
 */
char* getdriversname( int driver_number )
{
  static char *drivers_name[BGIPS+1] = {
    "DETECT",
    "BGI",
    "BGIS",
    "BGITGIF",
    "BGIPS",
  };

  switch( driver_number ){
  case DETECT:
  case BGI:
  case BGIS:
  case BGITGIF:
  case BGIPS:
    return( drivers_name[driver_number] );
    break;
  }
  return( NULL );
}
char* getdrivername( void )
{
  return( getdriversname( XBGI->gdriver ) );
}
int installuserdriver( char *name, int (*detect)( void ) )
{
  rferror( "installuserdriver()" );
  return( 0 );
}

/*
 * functions for graph mode
 */
int getmaxmode( void )
{
  return( BGICOLOR );
}
int getgraphmode( void )
{
  return( XBGI->gmode );
}
char* getmodename( int mode_number )
{
  static char *mode_name[BGICOLOR+1] = {
    "BGIMONO",
    "BGIGRAY",
    "BGICOLOR",
  };

  switch( mode_number ){
  case BGIMONO:
  case BGIGRAY:
  case BGICOLOR:
    return( mode_name[mode_number] );
    break;
  }
  return( NULL );
}
void getmoderange( int graphdriver, int *lomode, int *himode )
{
  nserror( "getmoderange()" );
}
void setgraphmode( int gmode )
{
  int i;
  unsigned long plane_mask;

  XBGI->gmode = gmode;
  switch( XBGI->gmode ){
  case BGIMONO:
    setmaxcolor( 2 );
    break;
  case BGIGRAY:
    setmaxcolor( XBGI->palette.size );
    break;
  case BGICOLOR:
    setmaxcolor( XBGI->palette.size );
    break;
  case NOWINDOWMODE:
    rferror( "setgraphmode( NOWINDOWMODE )" );
    setmaxcolor( XBGI->palette.size );
    break;
  }
}

/*
 * functions for color/palette
 */
int getmaxcolor( void )
{
  return( XBGI->palette.size-1 );
}
int getcolor( void )
{
  return( XBGI->color );
}
int getbkcolor( void )
{
  return( XBGI->bkcolor );
}
void setcolor( int color )
{
  XBGI->color = minmax( 0, color, XBGI->palette.size-1 );
  XSetForeground( XBGI->dpy, XBGI->gcdraw, XBGI->pixels[XBGI->color] );
}
void setbkcolor( int color )
{
  XBGI->bkcolor = minmax( 0, color, XBGI->palette.size-1 );
  XSetWindowBackground( XBGI->dpy, XBGI->win[XBGI->apage], 
                       XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcdraw,  XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcfill,  XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcpixel, XBGI->pixels[XBGI->bkcolor] );
  XSetBackground( XBGI->dpy, XBGI->gcimage, XBGI->pixels[XBGI->bkcolor] );
  /*XClearWindow( XBGI->dpy, XBGI->win[XBGI->apage] );*/
}
int getpalettesize( void )
{
  return( XBGI->palette.size );
}
struct palettetype* getdefaultpalette( void )
{
  return( &XBGI->palette );
}
void getpalette( struct palettetype *palette )
{
  *palette = XBGI->palette;
}
void setallpalette( struct palettetype *palette )
{
  int c;

  XBGI->palette = *palette;
  for ( c=0; c<palette->size; c++ )
    setpalette( palette->colors[c], c ); 
}
void setpalette( int colornum, int color )
{
  XColor xcolor;
  int red, green, blue;
  
  switch( XBGI->gmode ){
  case BGIMONO:
    break;
  case BGIGRAY:
  case BGICOLOR:
    colornum = minmax( 0, colornum, XBGI->palette.size-1 );
    color    = minmax( 0, color,    XBGI->palette.size-1 );
    XBGI->palette.colors[color] = (unsigned char)colornum;
    red   = XBGI->rgb[colornum].r;
    green = XBGI->rgb[colornum].g;
    blue  = XBGI->rgb[colornum].b;
    xcolor.pixel = XBGI->pixels[color];
    xcolor.red   = minmax( 0, 65536*red  /XBGI->palette.size, 65536-1 );
    xcolor.green = minmax( 0, 65536*green/XBGI->palette.size, 65536-1 );
    xcolor.blue  = minmax( 0, 65536*blue /XBGI->palette.size, 65536-1 );
    xcolor.flags = DoRed | DoGreen | DoBlue;
    XStoreColor( XBGI->dpy, XBGI->colormap, &xcolor );
    XFlush( XBGI->dpy );
    break;
  }
}
void setrgbpalette( int colornum, int red, int green, int blue )
{
  XColor xcolor;

  switch( XBGI->gmode ){
  case BGIMONO:
    break;
  case BGIGRAY:
  case BGICOLOR:
    colornum = minmax( 0, colornum, XBGI->palette.size-1 );
    XBGI->rgb[colornum].r = red;
    XBGI->rgb[colornum].g = green;
    XBGI->rgb[colornum].b = blue;
    xcolor.pixel = XBGI->pixels[colornum];
    xcolor.red   = minmax( 0, 65536*red  /XBGI->palette.size, 65536-1 );
    xcolor.green = minmax( 0, 65536*green/XBGI->palette.size, 65536-1 );
    xcolor.blue  = minmax( 0, 65536*blue /XBGI->palette.size, 65536-1 );
    xcolor.flags = DoRed | DoGreen | DoBlue;
    XStoreColor( XBGI->dpy, XBGI->colormap, &xcolor );
    XFlush( XBGI->dpy );
    break;
  }
}

/*
 * functions for fonts
 */
#define MAXFONT      4
#define MAXFONTSIZE 10
static char *FONTNAME[MAXFONT+1][MAXFONTSIZE+1] = {
#include "fonts/udef.chr"
#include "fonts/trip.chr"
#include "fonts/litt.chr"
#include "fonts/sans.chr"
#include "fonts/goth.chr"
};
static XFontStruct *FONTS[MAXFONT+1][MAXFONTSIZE+1];
void gettextsettings( struct textsettingstype *texttypeinfo )
{
  *texttypeinfo = XBGI->textset;
}
int installuserfont( char *name )
{
  rferror( "installuserfont()" );
  return( 0 );
}
void outtext( char *textstring )
{
  outtextxy( XBGI->cp.x, XBGI->cp.y, textstring );
  moverel( textwidth(textstring), 0 );
}
void outtextxy( int x, int y, char *textstring )
{
  int textx, texty, width, height;
  int i, j, k;
  Pixmap pixmap;
  XImage *image;
  unsigned char *data;
  GC gc;

  switch ( XBGI->textset.direction ){
  case VERT_DIR:
    width  = textwidth(textstring);
    height = textheight(textstring);
    pixmap = XCreatePixmap( XBGI->dpy, XBGI->win[XBGI->apage], width, height, XBGI->depth );
    gc = XCreateGC( XBGI->dpy, pixmap, 0, 0 );
    XSetForeground( XBGI->dpy, gc, XBGI->pixels[XBGI->bkcolor] );
    XFillRectangle( XBGI->dpy, pixmap, gc, 0, 0, width, height );
    XSetForeground( XBGI->dpy, gc, XBGI->pixels[XBGI->color] );
    XSetFont( XBGI->dpy, gc, XBGI->fsp->fid );
    XDrawString( XBGI->dpy, pixmap, gc, 0, XBGI->fsp->max_bounds.ascent,
                textstring, strlen(textstring) );
    switch( XBGI->textset.horiz ){
    case LEFT_TEXT:
      textx = x + XBGI->fsp->max_bounds.ascent;
      break;
    case CENTER_TEXT:
      textx = x + ( XBGI->fsp->max_bounds.ascent + XBGI->fsp->max_bounds.descent )/2;
      break;
    case RIGHT_TEXT:
      textx = x - XBGI->fsp->max_bounds.descent;
      break;
    }
    switch( XBGI->textset.vert ){
    case TOP_TEXT:
      texty = y;
      break;
    case CENTER_TEXT:
      texty = y - width/2;
      break;
    case BOTTOM_TEXT:
      texty = y - width;
      break;
    }
    image = XGetImage( XBGI->dpy, pixmap, 0, 0, width, height, AllPlanes, ZPixmap );
    data = (unsigned char *)image->data;
    for ( i=0; i<height; i++ ){
      for ( k=j=0; j<image->bytes_per_line; j++, data++ ){
	if ( j<image->xoffset ) continue;
	k++;
	if ( width < k ) continue;
	if ( XBGI->colortable[*data] == XBGI->color )
	  XDrawPoint( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
		     textx + i, texty + ( width - k ) );
      }
    }
    XDestroyImage( image );
    XFreeGC( XBGI->dpy, gc );
    XFreePixmap( XBGI->dpy, pixmap );
    break;
  case HORIZ_DIR:
    switch( XBGI->textset.horiz ){
    case LEFT_TEXT:
      textx = x;
      break;
    case CENTER_TEXT:
      textx = x - textwidth(textstring)/2;
      break;
    case RIGHT_TEXT:
      textx = x - textwidth(textstring);
      break;
    }
    switch( XBGI->textset.vert ){
    case TOP_TEXT:
      texty = y + XBGI->fsp->max_bounds.ascent;
      break;
    case CENTER_TEXT:
      texty = y + ( XBGI->fsp->max_bounds.ascent + XBGI->fsp->max_bounds.descent )/2;
      break;
    case BOTTOM_TEXT:
      texty = y - XBGI->fsp->max_bounds.descent;
      break;
    }
    XDrawString( XBGI->dpy, XBGI->win[XBGI->apage], XBGI->gcdraw,
                Xx( textx ), Xy( texty ), textstring, strlen(textstring) );
    break;
  }
}
void settextjustify( int horiz, int vert )
{
  XBGI->textset.horiz = minmax( LEFT_TEXT, horiz, RIGHT_TEXT );
  XBGI->textset.vert  = minmax( BOTTOM_TEXT, vert, TOP_TEXT );
}
void settextstyle( int font, int direction, int charsize )
{
  int i, j;
  static int flg = !0, flgs[MAXFONT+1][MAXFONTSIZE+1];

  if ( flg )
    for ( i=0; i<=MAXFONT; i++ )
      for ( j=0; j<=MAXFONTSIZE; j++ )
        flgs[i][j] = !0;
  flg = 0;
  XBGI->textset.font      = minmax( 0, font, MAXFONT );
  XBGI->textset.direction = minmax( 0, direction, 1 );
  XBGI->textset.charsize  = minmax( 0, charsize, MAXFONTSIZE );
  if ( flgs[XBGI->textset.font][XBGI->textset.charsize] )
    FONTS[XBGI->textset.font][XBGI->textset.charsize]
      = XLoadQueryFont( XBGI->dpy, FONTNAME[XBGI->textset.font][XBGI->textset.charsize] );
  XBGI->fsp = FONTS[XBGI->textset.font][XBGI->textset.charsize];
  XSetFont( XBGI->dpy, XBGI->gcdraw, XBGI->fsp->fid );
}
void setusercharsize( int multx, int divx, int multy, int divy )
{
  rferror( "setusercharsize()" );
}
int textheight( char *textstring )
{
  return( XBGI->fsp->max_bounds.ascent + XBGI->fsp->max_bounds.descent );
}
int textwidth( char *textstring )
{
  return( XTextWidth( XBGI->fsp, textstring, strlen(textstring)) );
}

/*
 * internal functions to initialize for xbgi
 */
void _initgraph_xbgi( int argc, char *argv[] )
{
  int i, flag_mask;
  Window parentwin;
  XSizeHints def;
  XSetWindowAttributes att;
  XEvent ev;
  char displayname[10] = "";
  char fnname[80]      = "lucidasans-bold-12";
  char fgname[20]      = "white";
  char bgname[20]      = "gray50";
  char lcname[20]      = "gray70";
  char dcname[20]      = "gray30";
  char defgeomstr[20]  = "640x400";
  char usrgeomstr[20]  = "";
  char titlename[80]   = "";
  char iconname[80]    = "";
  
  XBGI->ncolor  = 16;
  XBGI->color   = 15;
  XBGI->bkcolor = 0;
  XBGI->npage   = 2;
  XBGI->apage   = 0;
  XBGI->vpage   = 0;
  strcpy( titlename, argv[0] );
  strcpy( iconname,  argv[0] );

  /* set at exit procedure as "-atexit" */
  /*if ( scan_option( argc, argv, "-atexit" ) )
    atexit( xbgiloop );*/

  /* get display name and initialization */
  scan_option( argc, argv, "-display %s", displayname );
  XBGI->dpy = XOpenDisplay( displayname );
  XBGI->scr = DefaultScreen( XBGI->dpy );
  XBGI->colormap = DefaultColormap( XBGI->dpy, XBGI->scr );

  /* position, size, gc, font and color settings for work */
  scan_option( argc, argv, "-geometry %s", usrgeomstr );
  scan_option( argc, argv, "-fn %s", fnname );
  scan_option( argc, argv, "-fg %s", fgname );
  scan_option( argc, argv, "-bg %s", bgname );
  scan_option( argc, argv, "-lc %s", lcname );
  scan_option( argc, argv, "-dc %s", dcname );
  XBGI->gc = XCreateGC( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr), 0, NULL );
  XBGI->fs = XLoadQueryFont( XBGI->dpy, fnname );
  XBGI->fg = XGetPixelColorName( XBGI->dpy, fgname );
  XBGI->bg = XGetPixelColorName( XBGI->dpy, bgname );
  XBGI->lc = XGetPixelColorName( XBGI->dpy, lcname );
  XBGI->dc = XGetPixelColorName( XBGI->dpy, dcname );
  XSetForeground( XBGI->dpy, XBGI->gc, XBGI->fg );
  XSetFont( XBGI->dpy, XBGI->gc, XBGI->fs->fid );

  /* titlename and iconname settings */
  scan_option( argc, argv, "-title %s", titlename );
  scan_option( argc, argv, "-name  %s", iconname  );

  /* get user bgi settings */
  scan_option( argc, argv, "-gd %d", &(XBGI->gdriver) );
  scan_option( argc, argv, "-gm %d", &(XBGI->gmode) );
  scan_option( argc, argv, "-ncolor  %d", &(XBGI->ncolor) );
  scan_option( argc, argv, "-color   %d", &(XBGI->color) );
  scan_option( argc, argv, "-bkcolor %d", &(XBGI->bkcolor) );
  scan_option( argc, argv, "-npage %d", &(XBGI->npage) );
  scan_option( argc, argv, "-apage %d", &(XBGI->apage) );
  scan_option( argc, argv, "-vpage %d", &(XBGI->vpage) );
  XBGI->palette.size = (unsigned char)(XBGI->ncolor-1);

  /* create windows */
  flag_mask = XGeometry( XBGI->dpy, XBGI->scr, usrgeomstr, defgeomstr,
                        0, 1, 1, 0, 0, &def.x, &def.y, &def.width, &def.height );
  XBGI->root = XCreateSimpleWindow( XBGI->dpy, RootWindow(XBGI->dpy,XBGI->scr),
                                   def.x, def.y, def.width, def.height, 0,
                                   WhitePixel(XBGI->dpy,XBGI->scr),
                                   BlackPixel(XBGI->dpy,XBGI->scr) );
  if ( flag_mask&XValue )      def.flags = PPosition;
  if ( flag_mask&YValue )      def.flags = def.flags | USPosition;
  if ( flag_mask&WidthValue )  def.flags = def.flags | PSize;
  if ( flag_mask&HeightValue ) def.flags = def.flags | USSize;
  XSetStandardProperties( XBGI->dpy, XBGI->root, titlename, iconname, None, argv, argc, &def );
  
  /* input event selection, GC creation and so on */
  XBGI->depth  = DefaultDepth( XBGI->dpy, XBGI->scr );
  XBGI->gcdraw  = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcfill  = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcpixel = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  XBGI->gcimage = XCreateGC( XBGI->dpy, XBGI->root, 0, NULL );
  att.backing_store = Always;
  XBGI->win = (Window*)malloc( XBGI->npage*sizeof(Window) );
  for ( i=0; i<XBGI->npage; i++ ){
    XBGI->win[i] = XCreateSimpleWindow( XBGI->dpy, XBGI->root, 0, 0, def.width, def.height, 0,
                                       WhitePixel(XBGI->dpy,XBGI->scr),
                                       BlackPixel(XBGI->dpy,XBGI->scr) );
    XSelectInput( XBGI->dpy, XBGI->win[i], ExposureMask|
		 KeyPressMask|KeyReleaseMask|ButtonPressMask|ButtonReleaseMask|
		 Button1MotionMask|Button2MotionMask );
    XChangeWindowAttributes( XBGI->dpy, XBGI->win[i], CWBackingStore, &att );
    XSetStandardProperties( XBGI->dpy, XBGI->win[i],
			   titlename, iconname, None, argv, argc, &def );
  }
  
  /* BGI default setting */
  detectgraph( &(XBGI->gdriver), &(XBGI->gmode) );
  setgraphmode( XBGI->gmode );
  graphdefaults();
  
  XMapWindow( XBGI->dpy, XBGI->root );
  XMapSubwindows( XBGI->dpy, XBGI->root ); 
  for ( i=0; i<XBGI->npage; i++ )
    do {
      XNextEvent( XBGI->dpy, &ev );
      /*printf( "." );*/
    } while ( Expose != ev.type );
  /*printf( "\n" );*/
  XFlush( XBGI->dpy );
  XBGI->gerror = grOk;
}
void initgraph_xbgi( int *graphdriver, int *graphmode, char *pathtodriver )
{
  int i, argc = 0;
  char **argv, *ptr, options[256], buf[256];
  char *xbgienviron = "XBGISET";
  
  /*
   * parameter settings to argc, argv
   */
  sprintf( options, "%s -gd %d -gm %d ", "xbgi", *graphdriver, *graphmode );
  if ( getenv( xbgienviron ) != NULL ){
    strcat( options, getenv( xbgienviron ) );
    strcat( options, " " );
  }
  strcat( options, pathtodriver );
  strcat( options, " " );
  strcpy( buf, options );
  ptr = strtok( options, " " );
  while ( ptr != NULL ){
    argc++;
    ptr = strtok( NULL, " " );
  }
  argv = (char**)malloc( argc*sizeof(char*) );
  strcpy( options, buf );
  i = 0;
  ptr = strtok( options, " " );
  while ( ptr != NULL ){
    argv[i++] = ptr;
    ptr = strtok( NULL, " " );
  }
  _initgraph_xbgi( argc, argv );
  free( argv );
}

/*
 * other functions
 */
void _graphfreemem( void *ptr, unsigned size )
{
  nserror( "_graphfreemem()" );
}
void* _graphgetmem( unsigned size )
{
  nserror( "_graphgetmem()" );
  return( NULL );
}
void restorecrtmode( void )
{
  nserror( "restorecrtmode()" );
}
unsigned setgraphbufsize( unsigned bufsize )
{
  nserror( "setgraphbufsize()" );
  return( 0 );
}

/*
 * save to x bitmap file
 */
int savegraphtoxbm( char *filename )
{
  int apage = XBGI->apage;
  Pixmap bitmap;
  GC gc;
  XImage *image;
  int i, j, k, width, height;
  unsigned char *data;

  setactivepage( XBGI->vpage );
  width  = getmaxx()+1;
  height = getmaxy()+1;
  bitmap = XCreatePixmap( XBGI->dpy, XBGI->win[XBGI->apage], width, height, 1 );
  gc = XCreateGC( XBGI->dpy, bitmap, 0, 0 );
  XSetForeground( XBGI->dpy, gc, WhitePixel(XBGI->dpy,XBGI->scr) );
  XSetBackground( XBGI->dpy, gc, BlackPixel(XBGI->dpy,XBGI->scr) );
  XSetFillStyle( XBGI->dpy, gc, FillSolid );
  XFillRectangle( XBGI->dpy, bitmap, gc, 0, 0, width, height );
  XSetForeground( XBGI->dpy, gc, BlackPixel(XBGI->dpy,XBGI->scr) );
  XSetBackground( XBGI->dpy, gc, WhitePixel(XBGI->dpy,XBGI->scr) );
  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->apage],
                    0, 0, width, height, AllPlanes, ZPixmap );
  data = (unsigned char *)image->data;
  for ( i=0; i<height; i++ ){
    for ( k=j=0; j<image->bytes_per_line; j++, data++ ){
      if ( j<image->xoffset ) continue;
      k++;
      if ( width < k ) continue;
      if ( XBGI->colortable[*data] != XBGI->bkcolor )
	XDrawPoint( XBGI->dpy, bitmap, gc, k, i );
    }
  }
  XDestroyImage( image );
  XWriteBitmapFile( XBGI->dpy, filename, bitmap, width, height, 0, 0 );
  XFreeGC( XBGI->dpy, gc );
  XFreePixmap( XBGI->dpy, bitmap );
  setactivepage( apage );
}
/*
 * save to x pixmap file
 */
int savegraphtoxpm( char *filename )
{
  XImage *image;
  int apage = XBGI->apage;

#ifdef ENABLE_XPMLIB
  setactivepage( XBGI->vpage );
  image = XGetImage( XBGI->dpy, XBGI->win[XBGI->vpage],
                    0, 0, getmaxx()+1, getmaxy()+1, AllPlanes, ZPixmap );
  XpmWriteFileFromImage( XBGI->dpy, filename, image, 0, NULL );
  setactivepage( apage );
  XDestroyImage( image );
#endif /* ENABLE_XPMLIB */
}

/*
 * xbgiloop: standard loop with tools in XBGI
 */
/* menu defined */
enum MENU_ITEM {
  NO_SELECT = -1,
  SAVE_XBM,
  SAVE_XPM,
  QUIT,
};
static char *MENU_TITLE = "menu";
static int   MENU_ITEMC = 3;
static int   MENU_ITEMI = NO_SELECT;
static char *MENU_ITEMV[] = {
  "save bitmap",
#ifdef ENABLE_XPMLIB
  "save pixmap",
#else
  "*save pixmap",
#endif
  "quit (q)"
};

/* main loop with various tools */
void xbgiloop( void )
{
  xbgi *ptr;
  XEvent ev;
  char c;
  static char xbmfilename[80], xbmfilenamebak[80] = "scratch.xbm";
  static char xpmfilename[80], xpmfilenamebak[80] = "scratch.xpm";

  /*
   * flush all xbgi
   */
  ptr = XBGIS;
  while ( ptr != NULL ){
    XFlush( ptr->dpy );
    ptr = ptr->next;
  }
  ptr = XBGI;
  XBGI = XBGIS;
  /*
   * finite roop by quit(q/button3) and ctrl+c to abort
   */
  while ( !0 ){
    if ( XBGI == NULL ){
      XBGI = XBGIS;
      continue;
    }
    if ( XEventsQueued( XBGI->dpy, QueuedAfterFlush ) > 0 ){
      XNextEvent( XBGI->dpy, &ev );
      switch ( ev.type ){
      case KeyPress:
        if ( XLookupString( (XKeyEvent*)&ev, &c, 1, NULL, NULL ) == 1 ){
	  XBGI->key_press_ev = ev;
          switch( c ){
          case CTRL('a'):
            bgiprintf( "about", CENTER_TEXT, "xbgi\nversion 2\ngraphics.h for X\nTaiji Yamada" );
            break;
          case CTRL('i'):
            bgiprintf( "info.", LEFT_TEXT,
		      "graph id %d\ndriver %s\nmode %s\nvisual page %d/%d\nactive page %d/%d\n",
		      XBGI->id, getdriversname(XBGI->gdriver), getmodename(XBGI->gmode),
		      XBGI->apage, XBGI->npage, XBGI->vpage, XBGI->npage );
            break;
          case CTRL('b'):
	    strcpy( xbmfilename, xbmfilenamebak );
	    bgigets( "XBM NAME", 80, xbmfilename );
            if ( strcmp(xbmfilename,"") != 0 ){
	      savegraphtoxbm( xbmfilename );
	      strcpy( xbmfilenamebak, xbmfilename );
	    }else{
	      strcpy( xbmfilename, xbmfilenamebak );
            }
	    break;
          case CTRL('x'):
	    strcpy( xpmfilename, xpmfilenamebak );
	    bgigets( "XPM NAME", 80, xpmfilename );
            if ( strcmp(xpmfilename,"") != 0 ){
	      savegraphtoxpm( xpmfilename );
	      strcpy( xpmfilenamebak, xpmfilename );
	    }else{
	      strcpy( xpmfilename, xpmfilenamebak );
            }
	    break;
          case CTRL('t'):
            /*savegraphtoobj( objfilename );*/
            break;
          case CTRL('p'):
            /*savegraphtoeps( epsfilename );*/
            break;
          case 'n': case 'N':
            setvisualpage( min( XBGI->vpage+1, XBGI->npage-1 ) );
            break;
          case 'b': case 'B':
            setvisualpage( XBGI->vpage-1 );
            break;
          case CTRL('f'):
            if ( XBGI->vpage != XBGI->npage )
              setvisualpage( XBGI->npage );
            else
              setvisualpage( 0 );
            break;
          case CTRL('c'):
            exit( 0 );
            break;
          case 'q': case 'Q':
            XBGI = ptr;
            return;
            break;
          }
          break;
        }
        break;
      case KeyRelease:
        break;
      case ButtonPress:
        switch( ev.xbutton.button ){
        case Button1:
        case Button2:
	  XBGI->button_press_ev = ev;
          break;
        }
        switch( ev.xbutton.button ){
        case Button1:
          break;
        case Button2:
          bgimenu( MENU_TITLE, MENU_ITEMC, MENU_ITEMV, &MENU_ITEMI );
          switch( MENU_ITEMI ){
          case NO_SELECT:
            continue;
            break;
          case SAVE_XBM:
	    strcpy( xbmfilename, xbmfilenamebak );
	    bgigets( "XBM NAME", 80, xbmfilename );
            if ( strcmp(xbmfilename,"") != 0 ){
	      savegraphtoxbm( xbmfilename );
	      strcpy( xbmfilenamebak, xbmfilename );
	    }else{
	      strcpy( xbmfilename, xbmfilenamebak );
            }
	    break;
          case SAVE_XPM:
	    strcpy( xpmfilename, xpmfilenamebak );
	    bgigets( "XPM NAME", 80, xpmfilename );
            if ( strcmp(xpmfilename,"") != 0 ){
	      savegraphtoxpm( xpmfilename );
	      strcpy( xpmfilenamebak, xpmfilename );
	    }else{
	      strcpy( xpmfilename, xpmfilenamebak );
            }
	    break;
          case QUIT:
            XBGI = ptr;
            return;
            break;
          }
          break;
        case Button3:
          XBGI = ptr;
          return;
          break;
        }
        break;
      case ButtonRelease:
        switch( ev.xbutton.button ){
        case Button1:
        case Button2:
	  XBGI->button_release_ev = ev;
          break;
        }
	break;
      /***case ConfigureNotify:
	XResizeWindow( XBGI->dpy, XBGI->win[XBGI->vpage],
		      ev.xconfigure.width, ev.xconfigure.height );
	break;***/
      }
    }
    XBGI = XBGI->next;
  }
}
