/*
 * Source File < xbgiutl.c >
 *
 * XBGI Utilities Functions: Borland Graphics Interface for Xlib
 *
 * Copyleft (c) 1993 Tokyo Denki University, Taiji Yamada
 */
#include <malloc.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include "xbgiutl.h"
#include "_bgi.h"

#define HemBrim        (DecoHemBits+DecoBrimBits)
#define Hem(size)      (DecoHemBits+(size)+DecoHemBits)
#define Brim(size)     (DecoBrimBits+(size)+DecoBrimBits)
#define TextHeight(fs) (fs->max_bounds.ascent+fs->max_bounds.descent)
#define TextAscent(fs) (fs->max_bounds.ascent)

void xbgiGetSizeVPrintf( XFontStruct *fs, int *width, int *height,
                        char *fmt, va_list argptr )
{
  int i, maxwidth, c = 0; 
  char str[256], *ptr;
  
  vsprintf( str, fmt, argptr );
  strcat( str, "\n" );
  ptr = strtok( str, "\n" );
  while ( ptr != NULL ){
    c++;
    ptr = strtok( NULL, "\n" );
  }
  ptr = str;
  *height = TextHeight(fs);
  maxwidth = XTextWidth( fs, ptr, strlen(ptr) );
  for ( i=0; i<c; i++ ){
    maxwidth = max( maxwidth, XTextWidth( fs,ptr,strlen(ptr) ) );
    ptr = ptr + strlen(ptr);
    while ( ptr++ == NULL );
  }
  *width  = Brim(maxwidth);
  *height = Brim(*height)*c;
}
void xbgiGetSizePrintf( XFontStruct *fs, int *width, int *height, char *fmt, ... )
{
  va_list argptr;
  
  va_start( argptr, fmt );
  xbgiGetSizeVPrintf( fs, width, height, fmt, argptr );
}
void xbgiVPrintf( Display *display, Window win, GC gc, XFontStruct *fs,
                 int textjust, char *fmt, va_list argptr )
{
  Window root;
  int x, y, width, height, border, depth;
  int i, c = 0, textx, texty;
  char str[256], *ptr;
  
  vsprintf( str, fmt, argptr );
  strcat( str, "\n" );
  ptr = strtok( str, "\n" );
  while ( ptr != NULL ){
    c++;
    ptr = strtok( NULL, "\n" );
  }
  XGetGeometry( display, win, &root, &x, &y, &width, &height, &border, &depth );
  ptr = str;
  for ( i=0; i<c; i++ ){
    switch( textjust ){
    case LEFT_TEXT:
      textx = HemBrim;
      break;
    case CENTER_TEXT:
      textx = ( width - XTextWidth(fs,ptr,strlen(ptr)) )/2;
      break;
    case RIGHT_TEXT:
      textx = width - HemBrim - XTextWidth(fs,ptr,strlen(ptr));
      break;
    }
    texty = HemBrim + TextAscent(fs) + i*Brim(TextHeight(fs));
    XDrawString( display, win, gc, textx, texty, ptr, strlen(ptr) );
    ptr = ptr + strlen(ptr);
    while ( ptr++ == NULL );
  }
}
void xbgiPrintf( Display *display, Window win, GC gc, XFontStruct *fs,
                int textjust, char *fmt, ... )
{
  va_list argptr;

  va_start( argptr, fmt );
  xbgiVPrintf( display, win, gc, fs, textjust, fmt, argptr );
}
void xbgiDecorateWindow( Display *display, Window win, GC gc,
                        unsigned long fg, unsigned long lc, unsigned long dc, int type )
{
  XPoint orec[4], irec[4], lrec[4], trec[4], rrec[4], brec[4];
  XWindowAttributes att;

  XGetWindowAttributes( display, win, &att );
  orec[0].x = 0;                     orec[0].y = 0; 
  orec[1].x = 0;                     orec[1].y = att.height; 
  orec[2].x = att.width;             orec[2].y = att.height; 
  orec[3].x = att.width;             orec[3].y = 0; 
  irec[0].x = DecoHemBits;           irec[0].y = DecoHemBits; 
  irec[1].x = DecoHemBits;           irec[1].y = att.height-DecoHemBits; 
  irec[2].x = att.width-DecoHemBits; irec[2].y = att.height-DecoHemBits; 
  irec[3].x = att.width-DecoHemBits; irec[3].y = DecoHemBits; 
  lrec[0] = orec[0]; lrec[1] = orec[1]; lrec[2] = irec[1]; lrec[3] = irec[0];
  trec[0] = orec[0]; trec[1] = orec[3]; trec[2] = irec[3]; trec[3] = irec[0];
  rrec[0] = orec[2]; rrec[1] = orec[3]; rrec[2] = irec[3]; rrec[3] = irec[2];
  brec[0] = orec[1]; brec[1] = orec[2]; brec[2] = irec[2]; brec[3] = irec[1];
  XClearWindow( display, win );
  switch( type ){
  case DecoNothing:
    break;
  case DecoRectFront: 
    XSetForeground( display, gc, lc );
    XFillPolygon( display, win, gc, lrec, 4, Complex, CoordModeOrigin );
    XFillPolygon( display, win, gc, trec, 4, Complex, CoordModeOrigin );
    XSetForeground( display, gc, dc );
    XFillPolygon( display, win, gc, rrec, 4, Complex, CoordModeOrigin );
    XFillPolygon( display, win, gc, brec, 4, Complex, CoordModeOrigin );
    break;
  case DecoRectBack: 
    XSetForeground( display, gc, dc );
    XFillPolygon( display, win, gc, lrec, 4, Complex, CoordModeOrigin );
    XFillPolygon( display, win, gc, trec, 4, Complex, CoordModeOrigin );
    XSetForeground( display, gc, lc );
    XFillPolygon( display, win, gc, rrec, 4, Complex, CoordModeOrigin );
    XFillPolygon( display, win, gc, brec, 4, Complex, CoordModeOrigin );
    break;
  }
  XSetForeground( display, gc, fg );
}
